#include "stdafx.h"
#include "mIRCTileHoriz.h"

// Tell mIRC to load the DLL like this:
//
//   /dll "C:\Path to Dll\mIRCTileHoriz.dll" InstallHook /#topchannel [/#anotherchannel [/Status: /
//
// InstallHook must be specified, but everything afterwards is optional. The stuff
// afterwards is a list of window-title prefixes, separated by / characters, which should
// be kept in that whenever visible. Any other windows will be kept in the same order they
// are in already (i.e. Y-coordinate order), but that isn't completely reliable when windows
// are hidden/shown etc., so giving an explicit list of channel/window names to keep in
// order lets you ensure the order doesn't get messed up.
//
// You can also put **ignore** at the start of a prefix so that that window is completely ignored:
//
//   /dll "C:\Path to Dll\mIRCTileHoriz.dll" InstallHook /#topchannel [/#anotherchannel [/**ignore**Status: /
//
// Tell mIRC to unload the DLL like this:
//
//   /dll -u "C:\Path to Dll\mIRCTileHoriz.dll"
//
// You do not need to unload the DLL first if you just want to change the list of prefixes.
// mIRC will automatically unload the DLL before exiting, too.
// So the only times you'd need to unload the DLL manually is if you want to disable or replace it.

BOOL APIENTRY DllMain(HMODULE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	/*
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
	*/

	return TRUE;
}

bool g_bPendingLayout = false;
UINT g_uiLayoutMsg = 0;
UINT g_uiUnhookMsg = 0;
HWND g_hwndMdiChildHooked = NULL;

std::vector< std::wstring > g_vecTitlePrefixOrder;
std::vector< std::wstring > g_vecTitlePrefixIgnore;

struct LOADINFO
{
   DWORD  mVersion;
   HWND   mHwnd;
   BOOL   mKeep;
   BOOL   mUnicode;
};

LRESULT CALLBACK MdiChildSubclass(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, UINT_PTR uIdSubclass, DWORD_PTR dwRefData);
void UninstallHook();

inline bool IsUtf16LeftSurrogate(wchar_t c)
{
	// The casts might help if wchar_t is signed. :)
	return (((unsigned short)c) >= ((unsigned short)0xD800) && ((unsigned short)c) <= ((unsigned short)0xDBFF));
}

// Returns 0 on failure (use GetLastError()), else the number of wide characters written.
// delete [] *pwcStringResult when finished with it.
// numInputBytes -1 for null-terminated string.
// dwCodePage normally CP_ACP.
int MBtoWC(wchar_t **pwcStringResult, const char *mbStringInput, int numInputBytes, DWORD dwCodePage)
{
	*pwcStringResult = NULL;

	if (numInputBytes == 0)
	{
		*pwcStringResult = new(std::nothrow) wchar_t[1];

		if (NULL == *pwcStringResult)
		{
			::SetLastError(ERROR_NOT_ENOUGH_MEMORY);
			return 0;
		}

		(*pwcStringResult)[0] = L'\0';
		return 1;
	}

	int wcBufSize = ::MultiByteToWideChar(dwCodePage, 0, mbStringInput, numInputBytes, NULL, 0);

	if (wcBufSize == 0)
	{
		return 0;
	}

	if (numInputBytes != -1 && mbStringInput[numInputBytes-1] != '\0')
	{
		// If the string length is specified, and it doesn't contain a null, then MultiByteToWideChar will not include space for the null.
		// We'll add and allocate room for that extra null, whether the caller requires it or not. If the caller doesn't then it's harmless.
		++wcBufSize;
	}

	*pwcStringResult = new(std::nothrow) wchar_t[wcBufSize];

	if (NULL == *pwcStringResult)
	{
		::SetLastError(ERROR_NOT_ENOUGH_MEMORY);
		return 0;
	}

	int convRes = ::MultiByteToWideChar(dwCodePage, 0, mbStringInput, numInputBytes, *pwcStringResult, wcBufSize);

	(*pwcStringResult)[wcBufSize-1] = L'\0'; // No matter what, the buffer is terminated.

	if (convRes <= 0 || convRes > wcBufSize)
	{
//		assert(convRes==0); // Real WTF from the API if it isn't zero.

		delete[] *pwcStringResult;
		*pwcStringResult = NULL;

		return 0;
	}

	if (numInputBytes != -1 && mbStringInput[numInputBytes-1] != '\0' && convRes < wcBufSize)
	{
		++convRes; // Bump for the null that we wrote, which was not written by the API or included in its count.
	}

	// Finally, remove any left-surrogates from the end of the string.

	while(convRes > 1 && IsUtf16LeftSurrogate((*pwcStringResult)[convRes-2])) // -1 is the current null; -2 is the last character.
	{
		(*pwcStringResult)[convRes-2] = L'\0';
		--convRes;
	}

	// Ideally we'd ensure three are no partial multi-byte chars at the end of the string but I can't see a good way to do this
	// for the same dwCodePage argument which WideCharToMultiByte takes. CharPrevExA/CharNextExA are almost what we want, but not quite.

	return convRes;
}
void __stdcall LoadDll(LOADINFO* pInfo)
{
	g_uiUnhookMsg = ::RegisterWindowMessage(L"mIRCTileHoriz_1_{774E6733-984D-46e8-B9B7-B53C5B8AC076}");
	g_uiLayoutMsg = ::RegisterWindowMessage(L"mIRCTileHoriz_2_{774E6733-984D-46e8-B9B7-B53C5B8AC076}");
}

int __stdcall UnloadDll(int mTimeout)
{
	if (mTimeout == 1) // Unload request due to inactivity timeout.
	{
		return 0; // Tell mIRC to keep us loaded.
	}

	UninstallHook();

	return 1; // Tell mIRC to continue unloading us. (We don't have a choice in the matter unless mTimeout==1.)
}

int __stdcall InstallHook(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause)
{
	UninstallHook(); // Remove any existing hook.

	g_vecTitlePrefixOrder.clear();
	g_vecTitlePrefixIgnore.clear();

	const std::wstring strIgnorePrefix = L"**ignore**";

	if (data != NULL && data[0] != '\0')
	{
		wchar_t *szData = NULL;
		if (MBtoWC(&szData, data, -1, CP_ACP) > 0)
		{
			wchar_t *szPrefix = szData;
			wchar_t *szPEnd = szPrefix;

			bool bEnd = false;

			while (!bEnd)
			{
				if (*szPEnd == L'\0')
				{
					bEnd = true;
				}
				
				if (*szPEnd == L'\0' || *szPEnd == L'/')
				{
					if (szPrefix < szPEnd)
					{
						*szPEnd = L'\0';

						if (_wcsnicmp(szPrefix, strIgnorePrefix.c_str(), strIgnorePrefix.size()) == 0)
						{
							const wchar_t *szPrefixToIgnore = szPrefix + strIgnorePrefix.size(); 

							if (*szPrefixToIgnore != L'\0')
							{
								g_vecTitlePrefixIgnore.push_back(szPrefixToIgnore);
							}
						}
						else
						{
							g_vecTitlePrefixOrder.push_back(szPrefix);
						}
					}

					++szPEnd;
					szPrefix = szPEnd;
				}
				else
				{
					++szPEnd;
				}
			}

			delete[] szData;
		}
	}

	if (g_uiUnhookMsg != 0
	&&	g_uiLayoutMsg != 0
	&&	g_hwndMdiChildHooked == NULL	// Avoid installing more than once. (In case UninstallHook fails, I guess.)
	&&	mWnd != NULL)					// Main window must exist so we can look for its children.
	{
		MircWindows mw(mWnd, NULL);

		HWND hwndMdiChild = mw.GetMdiChild();

		if (hwndMdiChild != NULL // MDIChild window must exist so we can hook it.
		&&	::GetWindowThreadProcessId(hwndMdiChild, NULL) == ::GetCurrentThreadId()) // MDIChild window must belong to this thread so we can hook it.
		{
			g_bPendingLayout = false;
			if (::SetWindowSubclass(hwndMdiChild, MdiChildSubclass, 0, 0))
			{
				g_hwndMdiChildHooked = hwndMdiChild;

				mw.DoLayout(g_vecTitlePrefixOrder, g_vecTitlePrefixIgnore); // Do initial layout, too.
			}
		}
	}

	return 1; // tell mIRC to continue
}


void UninstallHook()
{
	if (g_uiUnhookMsg == 0
	||	g_uiLayoutMsg == 0
	||	g_hwndMdiChildHooked == NULL)
	{
		// Hook isn't installed.
		return;
	}

	if (::GetWindowThreadProcessId(g_hwndMdiChildHooked, NULL) != ::GetCurrentThreadId())
	{
		// We're not on the window's thread. Send it a message telling it to unhook itself.
		::SendMessage(g_hwndMdiChildHooked, g_uiUnhookMsg, 0, 0);
		return;
	}

	::RemoveWindowSubclass(g_hwndMdiChildHooked, MdiChildSubclass, 0);
	g_hwndMdiChildHooked = NULL;
	g_bPendingLayout = false;
}

LRESULT CALLBACK MdiChildSubclass(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, UINT_PTR uIdSubclass, DWORD_PTR dwRefData)
{
	if ((g_uiUnhookMsg != 0 && uMsg == g_uiUnhookMsg)
	||	uMsg == WM_DESTROY)
	{
		UninstallHook();
	}
	else if (uMsg == WM_SIZE || uMsg == WM_IME_SETCONTEXT) // WM_IME_SETCONTEXT seems a reliable way to know when a channel/status window has been opened or closed.
	{
		if (!g_bPendingLayout)
		{
			::PostMessage(hWnd, g_uiLayoutMsg, 0, 0);
			g_bPendingLayout = true;
		}
	}
	else if (uMsg == g_uiLayoutMsg)
	{
		MircWindows mw(NULL, hWnd);
		mw.DoLayout(g_vecTitlePrefixOrder, g_vecTitlePrefixIgnore);

		g_bPendingLayout = false;
	}

	return DefSubclassProc(hWnd, uMsg, wParam, lParam);
}
