#include "stdafx.h"
#include "mIRCTileHoriz.h"

// See comments at the top of dllmain.cpp for usage within mIRC.

MircWindows::MircWindows(HWND hwndMain, HWND hwndMdiChild)
: m_hwndMain(hwndMain)
, m_hwndMdiChild(hwndMdiChild)
, m_hwndMaxChild(NULL)
, m_pvecTitlePrefixOrder(NULL)
, m_pvecTitlePrefixIgnore(NULL)
{
	if (m_hwndMain != NULL && m_hwndMdiChild == NULL)
	{
		::EnumChildWindows(m_hwndMain, EnumMdiChild, reinterpret_cast< LPARAM >( this ));
	}
	else if (m_hwndMain == NULL && m_hwndMdiChild != NULL)
	{
		m_hwndMain = ::GetParent(m_hwndMdiChild);
	}
}

bool MircWindows::DoLayout(const std::vector< std::wstring > &vecTitlePrefixOrder, const std::vector< std::wstring > &vecTitlePrefixIgnored)
{
	if (m_hwndMdiChild == NULL)
	{
		return false;
	}

	m_pvecTitlePrefixOrder = &vecTitlePrefixOrder;
	m_pvecTitlePrefixIgnore = &vecTitlePrefixIgnored;
	m_hwndMaxChild = NULL;

	::EnumChildWindows(m_hwndMdiChild, MircChannelDeMax, reinterpret_cast< LPARAM >( this ));

	if (m_hwndMaxChild != NULL)
	{
		::EnumChildWindows(m_hwndMdiChild, MircChannelMinOthers, reinterpret_cast< LPARAM >( this ));
	}

	::EnumChildWindows(m_hwndMdiChild, MircChannelEnum, reinterpret_cast< LPARAM >( this ));

	m_pvecTitlePrefixOrder = NULL;
	m_pvecTitlePrefixIgnore = NULL;
	m_hwndMaxChild = NULL;

	if (m_mmapChannels.empty())
	{
		return false;
	}

	int fx = ::GetSystemMetrics(SM_CXSIZEFRAME);
	int fy = ::GetSystemMetrics(SM_CYSIZEFRAME) - 1;

	RECT rcClient = {0};
	
	if (!::GetClientRect(m_hwndMdiChild, &rcClient))
	{
		return false;
	}

	rcClient.left   -= fx;
	rcClient.right  += fx;

	rcClient.top    -= 1;
	rcClient.bottom += fy;

	size_t i = 0;
	for (std::multimap< std::pair< LONG, LONG > , HWND >::const_iterator miter = m_mmapChannels.begin(); miter != m_mmapChannels.end(); ++miter)
	{
		RECT rcChild = rcClient;

		rcChild.top    = static_cast< LONG >( rcClient.top + ((rcClient.bottom - rcClient.top) *   i) / m_mmapChannels.size() ) - fy;
		rcChild.bottom = static_cast< LONG >( rcClient.top + ((rcClient.bottom - rcClient.top) * ++i) / m_mmapChannels.size() );

		::SetWindowPos(miter->second, NULL, rcChild.left, rcChild.top, rcChild.right - rcChild.left, rcChild.bottom - rcChild.top, SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER);
	}

	return true;
}

BOOL CALLBACK MircWindows::EnumMdiChild(HWND hWnd, LPARAM lParam)
{
	MircWindows *pThis = reinterpret_cast< MircWindows * >( lParam );

	wchar_t szClass[256];

	if (::GetClassName(hWnd, szClass, _countof(szClass))
	&&	::GetParent(hWnd) == pThis->m_hwndMain // Only want actual children. (EnumChildWindows is a badly named API.)
	&&	0 == _wcsicmp(szClass, L"MDIClient"))
	{
		pThis->m_hwndMdiChild = hWnd;
		return FALSE;
	}

	return TRUE;
}

BOOL CALLBACK MircWindows::MircChannelDeMax(HWND hWnd, LPARAM lParam)
{
	MircWindows *pThis = reinterpret_cast< MircWindows * >( lParam );

	if (::GetParent(hWnd) == pThis->m_hwndMdiChild // Only want actual children. (EnumChildWindows is a badly named API.)
	&&	(::GetWindowLong(hWnd, GWL_STYLE)&WS_MAXIMIZE))
	{
		if (pThis->m_hwndMaxChild == NULL)
		{
			pThis->m_hwndMaxChild = hWnd;
		}
		::ShowWindow(hWnd, SW_SHOWNOACTIVATE);
	}

	return TRUE;
}

BOOL CALLBACK MircWindows::MircChannelMinOthers(HWND hWnd, LPARAM lParam)
{
	MircWindows *pThis = reinterpret_cast< MircWindows * >( lParam );

	if (hWnd != pThis->m_hwndMaxChild
	&&	::IsWindowVisible(hWnd)
	&&	::GetParent(hWnd) == pThis->m_hwndMdiChild) // Only want actual children. (EnumChildWindows is a badly named API.)
	{
		if (!pThis->CheckPrefixes(hWnd, pThis->m_pvecTitlePrefixIgnore, NULL))
		{
			::ShowWindow(hWnd, SW_HIDE);
		}
	}

	return TRUE;
}

BOOL CALLBACK MircWindows::MircChannelEnum(HWND hWnd, LPARAM lParam)
{
	MircWindows *pThis = reinterpret_cast< MircWindows * >( lParam );

	RECT rc = {0};

	if (::IsWindowVisible(hWnd)
	&&	::GetParent(hWnd) == pThis->m_hwndMdiChild // Only want actual children. (EnumChildWindows is a badly named API.)
	&&	(::GetWindowLong(hWnd, GWL_STYLE)&WS_THICKFRAME)
	&&	::GetWindowRect(hWnd, &rc))
	{
		if (!pThis->CheckPrefixes(hWnd, pThis->m_pvecTitlePrefixIgnore, NULL))
		{
			LONG lSortWeight = 0;
			pThis->CheckPrefixes(hWnd, pThis->m_pvecTitlePrefixOrder, &lSortWeight);

			pThis->m_mmapChannels.insert( std::make_pair( std::make_pair(lSortWeight, rc.top) , hWnd ) );
		}
	}

	return TRUE;
}

bool MircWindows::CheckPrefixes(HWND hWnd, const std::vector< std::wstring > *pvecPrefixes, LONG *plSortWeight)
{
	if (pvecPrefixes == NULL)
	{
		if (plSortWeight != NULL)
		{
			*plSortWeight = 0;
		}
		return false;
	}

	if (plSortWeight != NULL)
	{
		*plSortWeight = pvecPrefixes->size();
	}

	wchar_t szTitle[1024];

	if (::GetWindowThreadProcessId(hWnd, NULL) != ::GetCurrentThreadId() // Shouldn't be an issue, but just in case it might happen and cause a deadlock.
	||	::GetWindowText(hWnd, szTitle, _countof(szTitle)) <= 0)
	{
		return false;
	}

	for (size_t i = 0; i < pvecPrefixes->size(); ++i)
	{
		if (0 == _wcsnicmp(szTitle, (*pvecPrefixes)[i].c_str(), (*pvecPrefixes)[i].size()))
		{
			if (plSortWeight != NULL)
			{
				*plSortWeight = static_cast< LONG >( i );
			}
			return true;
		}
	}

	return false;
}
